/* 
 * Copyright (C) 2004-2005 Jonathan Bindel
 * Copyright (C) 2006-2007 Eskil Bylund
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

using System;
using System.Diagnostics;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace DCSharp.Backend.Connections
{
	public abstract class Server
	{
		private IPAddress localIP;
		private int port;
		private Socket listener;
		private bool started;

		private readonly object syncRoot;

		#region Constructors

		protected Server(IPAddress localIP, int port)
		{
			if (localIP == null)
			{
				throw new ArgumentNullException("localIP");
			}
			if (port <= 0 || port > IPEndPoint.MaxPort)
			{
				throw new ArgumentOutOfRangeException("Port");
			}

			syncRoot = new object();

			this.localIP = localIP;
			this.port = port;
		}

		#endregion

		#region Properties

		public IPAddress LocalIP
		{
			get { return localIP; }
		}

		public int Port
		{
			get { return port; }
		}

		public Socket Listener
		{
			get { return listener; }
			protected set { listener = value; }
		}

		public bool IsStarted
		{
			get { return started; }
		}

		#endregion

		#region Methods

		public virtual void Start()
		{
			lock (syncRoot)
			{
				if (!started)
				{
					started = true;
					ServerStart();
				}
			}
		}

		public virtual void Stop()
		{
			lock (syncRoot)
			{
				if (!started)
				{
					return;
				}
				started = false;

				if (listener != null)
				{
					if (listener.Connected)
					{
						listener.Shutdown(SocketShutdown.Both);
					}
					listener.Close();
				}
			}
		}

		protected abstract void ServerStart();

		#endregion
	}
}
