/*
 * Copyright (C) 2007 Eskil Bylund
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

using System;
using System.Diagnostics;

namespace DCSharp.Logging
{
	/// <summary>
	/// A logging class.
	/// </summary>
	public class Logger
	{
		private string name;
		private TraceLevel level;

		/// <summary>
		/// Constructs a new logger instance.
		/// </summary>
		/// <param name="name">The name of the logger.</param>
		public Logger(string name)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			this.name = name;
			level = TraceLevel.Info;
		}

		/// <summary>
		/// Gets or sets the log level.
		/// </summary>
		/// <value>The log level.</value>
		public TraceLevel Level
		{
			get { return level; }
			set { level = value; }
		}

		/// <summary>
		/// Gets the name of the logger.
		/// </summary>
		/// <value>The name of the logger.</value>
		public string Name
		{
			get { return name; }
		}

		/// <summary>
		/// Logs a debug message.
		/// </summary>
		/// <param name="message">The message to log.</param>
		public void Debug(string message)
		{
			Debug(message, null);
		}

		/// <summary>
		/// Logs a debug message.
		/// </summary>
		/// <param name="message">The message to log.</param>
		/// <param name="e">The exception to log.</param>
		/// <param name="context">Context to the log message.</param>
		public void Debug(string message, Exception e, params object[] context)
		{
			Log(TraceLevel.Verbose, message, e, context);
		}

		/// <summary>
		/// Logs an error message.
		/// </summary>
		/// <param name="message">The message to log.</param>
		public void Error(string message)
		{
			Error(message, null);
		}

		/// <summary>
		/// Logs an error message.
		/// </summary>
		/// <param name="message">The message to log.</param>
		/// <param name="e">The exception to log.</param>
		/// <param name="context">Context to the log message.</param>
		public void Error(string message, Exception e, params object[] context)
		{
			Log(TraceLevel.Error, message, e, context);
		}

		/// <summary>
		/// Logs an info message.
		/// </summary>
		/// <param name="message">The message to log.</param>
		public void Info(string message)
		{
			Info(message, null);
		}

		/// <summary>
		/// Logs an info message.
		/// </summary>
		/// <param name="message">The message to log.</param>
		/// <param name="e">The exception to log.</param>
		/// <param name="context">Context to the log message.</param>
		public void Info(string message, Exception e, params object[] context)
		{
			Log(TraceLevel.Info, message, e, context);
		}

		/// <summary>
		/// Logs a message.
		/// </summary>
		/// <param name="level">The message log level.</param>
		/// <param name="message">The message to log.</param>
		/// <param name="e">The exception to log.</param>
		/// <param name="context">Context to the log message.</param>
		/// <remarks>
		/// The message is not logged if the message log level is above the log
		/// level of the logger.
		/// </remarks>
		protected void Log(TraceLevel level, string message, Exception e,
			params object[] context)
		{
			if (level <= Level)
			{
				LogManager.Log(this, level, message, e, context);
			}
		}
	}
}
