/* 
 * Copyright (C) 2004-2005 Jonathan Bindel, jonathanbindel@gmail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

using System;
using System.Collections.Generic;
using System.Xml.Serialization;

using DCSharp.Backend.Objects;
using DCSharp.Backend.Managers;
using DCSharp.Xml;

namespace DCSharp.Settings
{
	[XmlRoot("Settings")]
	public class RuntimeSettings : IRuntimeSettings
	{
		private UserSettings userSettings;
		private ConnectionSettings connectionSettings;
		private List<ShareInfo> shared;

		private string downloadDirectory;
		private string tempDownloadDirectory;
		private string protocolCodePage;
		private bool segmentedDownloading;
		private int? downloadSlots;
		private int? uploadSlots;
		private bool autoFollowRedirects;
		private bool shareHidden;

		public RuntimeSettings() : base()
		{
			connectionSettings = new ConnectionSettings();
			userSettings = new UserSettings();
			shared = new List<ShareInfo>();

			// Defaults
			downloadDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Personal);
			tempDownloadDirectory = downloadDirectory;
			protocolCodePage = "windows-1252";
			segmentedDownloading = true;
			downloadSlots = 6;
			uploadSlots = 6;
			autoFollowRedirects = true;
		}

		[XmlElement("User")]
		public UserSettings UserSettings
		{
			get { return userSettings; }
			set { userSettings = value; }
		}

		[XmlElement("Connection")]
		public ConnectionSettings ConnectionSettings
		{
			get { return connectionSettings; }
			set { connectionSettings = value; }
		}

		[XmlArray("Share")]
		[XmlArrayItem("Directory")]
		public List<ShareInfo> SharedDirectories
		{
			get { return shared; }
			set { shared = value; }
		}

		#region IPublicSettings

		[XmlIgnore]
		public LocalIdentity LocalIdentity
		{
			get { return userSettings.LocalIdentity; }
			set { userSettings.LocalIdentity = value; }
		}

		/// <summary>
		/// Gets or sets the download directory.
		/// </summary>
		public string DownloadDirectory
		{
			get { return downloadDirectory; }
			set { downloadDirectory = value; }
		}

		/// <summary>
		/// Gets or sets temporary directory for downloads.
		/// </summary>
		public string TempDownloadDirectory
		{
			get { return tempDownloadDirectory; }
			set { tempDownloadDirectory = value; }
		}

		/// <summary>
		/// Gets or sets the default protocol code page.
		/// </summary>
		public string ProtocolCodePage
		{
			get { return protocolCodePage; }
			set { protocolCodePage = value; }
		}

		/// <summary>
		/// Gets or sets whether or not to use segmented downloading.
		/// </summary>
		public bool SegmentedDownloading
		{
			get { return segmentedDownloading; }
			set { segmentedDownloading = value; }
		}

		/// <summary>
		/// Gets or sets the number of download slots.
		/// </summary>
		/// <value></value>
		public int? DownloadSlots
		{
			get { return downloadSlots; }
			set { downloadSlots = value; }
		}

		/// <summary>
		/// Gets or sets the number of upload slots.
		/// </summary>
		public int? UploadSlots
		{
			get { return uploadSlots; }
			set { uploadSlots = value; }
		}

		/// <summary>
		/// Gets or sets whether or not redirects should be automatically followed.
		/// </summary>
		/// <value></value>
		public bool AutoFollowRedirects
		{
			get { return autoFollowRedirects; }
			set { autoFollowRedirects = value; }
		}

		/// <summary>
		/// Gets or sets whether or not to share hidden files.
		/// </summary>
		/// <value></value>
		public bool ShareHidden
		{
			get { return shareHidden; }
			set { shareHidden = value; }
		}

		#endregion

		public void Save(string filename)
		{
			XmlHelper.Serialize(this, filename);
		}

		public static RuntimeSettings Load(string filename)
		{
			return (RuntimeSettings)XmlHelper.Deserialize(typeof(RuntimeSettings),
				filename);
		}
	}
}
