/*
 * Copyright (C) 2006-2007 Eskil Bylund
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

using System;
using System.Collections;
using System.Diagnostics;
using System.Text;

using Gtk;
using Glade;
using Mono.Unix;

using DCSharp.Backend.Connections;
using DCSharp.Backend.Objects;
using DCSharp.Backend.Managers;

namespace DCSharp.Gui
{
	public class ChatPage : Page
	{
		private User localUser;
		private User remoteUser;

		[Widget]
		private ChatView chatView;

		[Widget]
		private Entry messageEntry;

		#region Constructors

		public ChatPage(User localUser, User remoteUser) :
			base("ChatPage.glade")
		{
			if (localUser == null)
			{
				throw new ArgumentNullException("localUser");
			}
			if (remoteUser == null)
			{
				throw new ArgumentNullException("remoteUser");
			}

			this.localUser = localUser;
			this.remoteUser = remoteUser;

			chatView.Show();

			messageEntry.GrabFocus();
			messageEntry.Activated += OnMessageEntryActivated;

			// Events
			remoteUser.OnlineChanged += OnOnlineChanged;
		}

		#endregion

		#region Properties

		public override string Title
		{
			get { return remoteUser.Nick; }
		}

		public override Gdk.Pixbuf Icon
		{
			get { return Util.GetUserIcon(remoteUser); }
		}

		public override string Tooltip
		{
			get { return remoteUser.Nick; }
		}

		public User LocalUser
		{
			get { return localUser; }
		}

		public User RemoteUser
		{
			get { return remoteUser; }
		}

		#endregion

		#region Methods

		public void DisplayMessage(User from, string message)
		{
			if (from == null)
			{
				throw new ArgumentNullException("from");
			}
			if (message == null)
			{
				throw new ArgumentNullException("message");
			}
			if (from != localUser && from != remoteUser)
			{
				throw new ArgumentException("Unknown user.", "from");
			}

			chatView.InsertMessage(from.Nick, message, from == localUser);
			OnPageChanged();
		}

		protected override void OnDestroyed()
		{
			remoteUser.OnlineChanged -= OnOnlineChanged;
			base.OnDestroyed();
		}

		private void DisplayStatus()
		{
			string status = remoteUser.IsOnline ?
				Catalog.GetString("{0} has come online") :
				Catalog.GetString("{0} went offline");

			status = String.Format(status, remoteUser.Nick);

			chatView.InsertStatus(status);
			Status = status;
		}

		private void OnMessageEntryActivated(object o, EventArgs args)
		{
			string message = messageEntry.Text.Trim();
			if (message.Length > 0)
			{
				HubConnection sentTo = Globals.Runtime.HubManager.Find(
					delegate(HubConnection hub)
				{
					try
					{
						hub.SendPrivateMessage(remoteUser, message);
						messageEntry.Text = String.Empty;
						return true;
					}
					catch
					{
					}
					return false;
				});
				if (sentTo == null)
				{
					chatView.InsertError(
						Catalog.GetString("Could not send the message through any of the hubs."));
				}
			}
		}

		private void OnOnlineChanged(object obj, EventArgs args)
		{
			Application.Invoke(delegate
			{
				DisplayStatus();
				OnPageChanged();
			});
		}

		#endregion
	}
}
